#!/usr/bin/env python
#-*- coding: utf8 -*-
#
############################################################################
#    Copyright (C) 2010 by Daniel 'Beorn' Mróz <beorn@cyberdeck.pl>        #
#                                                                          #
#    This program is free software; you can redistribute it and/or modify  #
#    it under the terms of the GNU Library General Public License ver. 3.0 #
#    published by the Free Software Foundation.                            #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU Library General Public     #
#    License along with this program; if not, write to the                 #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################


import socket
import struct



class Address(object):
    """ A class representing a single IP address """
    def __init__(self, value):
        __slots__ = ['numeric', 'literal']
        """ Create an Address object
        
            ex:
                
                >>> # Create an object from string
                >>> Address('127.0.0.1')
                Address('127.0.0.1')
                
                >>> # Create an object from long
                >>> Address(2130706433)
                Address('127.0.0.1')
                
                >>> # Make a copy of an existing object
                >>> Address(Address('127.0.0.1')
                Address('127.0.0.1')
                
            Address object contains two attributes:
            
            Address.numeric - numeric representation of an IP address
            Address.literal - literal representation of an IP address
            
            Those attributes can be used to set a new value.
        """
        if isinstance(value, (str, unicode)):
            self.numeric = self._parse_string(str(value))
        elif isinstance(value, (int, long)):
            self.numeric = long(value)
        elif isinstance(value, Address):
            self.numeric = value.numeric
        else:
            raise ValueError('Invalid IP address type')
        
        
    def is_private(self):
        """ Return True if address belongs to one of the classful ranges: A, C, reseved
            localhost class (127.0.0.0/8), link-local (169.255.0.0/16) or CIDR 172.16/12 range.
            WARNING: this method DOES NOT check against multicast (class D) or
            reserved class E networks. If you need such test, you must do it
            manually (you can use 'multicast' and 'class_e' module attributes).
        """
        return (self in class_a) or (self in class_c) or (self in loopback) or (self in link_local)
        

    def is_loopback(self):
        """ Return True if current address belongs to loopback address range (127.0.0.0/8)
        """
        return self in loopback
    
    
    def is_link_local(self):
        """ Return True if current address belongs to link-local address range (169.254.0.0/16),
            which is used by some operating systems to assign an IP address to device, when
            there's is no configuration source or source is not available (i.e. configuration
            via DHCP fails).
        """
        return self in link_local
    
    
    def to_str(self):
        """ Return literal representation of an address.
            Same as Address.literal or str(Address)
        """
        return self.literal
    
    
    def to_hex(self):
        """ Return hexadecimal (as str type) representation of an address """
        return hex(self.numeric)[2:]
    
    
    def to_int(self):
        """ Return numeric (as int type) representation of an address.
            Same as Address.numeric or int(Address)
        """
        return self.numeric
    
    
    def to_bytes(self):
        """ Return byte representation of an address """
        return struct.pack('!L', long(self.numeric))
    
    
    @staticmethod
    def ston(address):
        """ Convert string IP address into long value """
        try:
            chunks = address.split('.')
            if len(chunks) != 4:
                raise ValueError
            for v in chunks:
                i = int(v)
                if i < 0 or i > 255:
                    raise ValueError
        except:
            raise ValueError('Invalid IP address and/or format: "{0}"'.format(address))
        
        return struct.unpack('!L', socket.inet_aton(str(address)))[0]
            
    
    @staticmethod
    def ntos(value):
        """ Convert long IP address value into string """
        if not Address.in_bounds(value):
            raise ValueError('Value out of bounds')
        return socket.inet_ntoa(struct.pack('!L', value))


    @staticmethod
    def in_bounds(value):
        """ Test for value boundary.
            Returns 'True' if 0 <= value <= 0xFFFFFFFF, 'False' otherwise.
        """
        if value <= 0xffffffff and value >= 0:
            return True
        return False
    

    def _parse_string(self, value):
        """ Parse literal representation of an IP address and return a long value """
        return self.ston(value)
    
    
    def __setattr__(self, name, value):
        if name == 'numeric':
            if not isinstance(value, (int, long)):
                raise AttributeError('Attribute \'{0}\' should be \'int\' or \'long\', not \'{1}\''.format(name, value.__class__.__name__))
            elif not self.in_bounds(value):
                raise ValueError('Value out of bounds')
            self.__dict__['_value'] = int(value)
        elif name == 'literal':
            if not isinstance(value, (str, unicode)):
                raise AttributeError('Attribute \'{0}\' should be \'str\' or \'unicode\', not \'{1}\''.format(name, value.__class__.__name__))
            self.__dict__['_value'] = self._parse_string(str(value))
        else:
            raise AttributeError('\'{0}\' object has no attribute \'{1}\''.format(self.__class__.__name__, name))
        
        
    def __getattr__(self, name):
        if name == 'numeric':
            return self.__dict__['_value']
        elif name == 'literal':
            return self.ntos(self.__dict__['_value'])
        else:
            raise AttributeError('\'{0}\' object has no attribute \'{1}\''.format(self.__class__.__name__, name))

        
    def __str__(self):
        return self.literal
    
    
    def __int__(self):
        return int(self.numeric)
    
    
    def __long__(self):
        return long(self.numeric)
    
    
    def __add__(self, value):
        new_value = self.numeric + long(value)
        if not self.in_bounds(new_value):
            raise ValueError('Value out of bounds after addition')
        return self.__class__(new_value)
    
    
    def __iadd__(self, value):
        new_value = self.numeric + long(value)
        if not self.in_bounds(new_value):
            raise ValueError('Value out of bounds after addition')
        self.numeric = new_value
        return self
    
    
    def __sub__(self, value):
        new_value = self.numeric - long(value)
        if not self.in_bounds(new_value):
            raise ValueError('Value out of bounds after substraction')
        return self.__class__(new_value)
    
    
    def __isub__(self, value):
        new_value = self.numeric - long(value)
        if not self.in_bounds(new_value):
            raise ValueError('Value out of bounds after addition')
        self.numeric = new_value
        return self
    
    
    def __eq__(self, other):
        if not isinstance(other, Address):
            raise ValueError('Cannot compare \'{0}\' with \'{1}\''.format(self.__class__.__name__, other.__class__.__name__))
        return self.numeric == other.numeric
    
    
    def __ne__(self, other):
        return not self.__eq__(other)
    
    
    def __cmp__(self, other):
        if not isinstance(other, Address):
            raise ValueError('Cannot compare \'{0}\' with \'{1}\''.format(self.__class__.__name__, other.__class__.__name__))
        return cmp(self.numeric, other.numeric)
    
        
    def __repr__(self):
        return '{0}(\'{1}\')'.format(self.__class__.__name__, self.literal)
        

    
        
class Network(object):
    """ Class representing a range of Address addresses. CIDR ranges can be used. """
    def __init__(self, network):
        __slots__ = ['first', 'last', 'mask', 'size']
        """ Create a Network object.
            
            ex:
            
                >>> Network('127.0.0.1 - 127.0.0.6')
                Network('127.0.0.1 - 127.0.0.6')
                
                >>> Network('127.0.0.0/8')
                Network('127.0.0.0 - 127.255.255.255')
                
                >>> Network((Address('127.0.0.0'), Address('127.255.255.255')))
                Network('127.0.0.0 - 127.255.255.255')
                
                >>> Network(Network('127.0.0.0/8'))
                Network('127.0.0.0 - 127.255.255.255')
                
            For strings other than CIDR ('XXX.XXX.XXX.XXX/YY') mask value will be
            calculated using method Network.calcmask() and may not be accurate.
        """
        if isinstance(network, (str, unicode)):
            # 'first - last' or 'first-last'
            (self.first, self.last, self.mask) = self._parse_string(network)
        elif isinstance(network, (list, tuple)):
            # (first, last)
            if len(network) != 2:
                raise ValueError('Invalid network value')            
            (self.first, self.last) = map(Address, network)
            self.mask = None   # calculate
        elif isinstance(network, Network):
            self.first = network.first
            self.last = network.last
            self.mask = network.mask
        else:
            self.first = Address(network)
            self.last = Address(self.first)
            self.mask = 32
        
        if self.last.numeric < self.first.numeric:
            raise ValueError('Range is reversed')
        
        # Reset iteration index
        self.__dict__['__iter_idx'] = None
        self.reset()
            
    
    def is_private(self):
        """ Return True if network or any of it's part belongs to one of the classful
            ranges: A, C, reseved localhost class (127.0.0.0/8), link-local (169.255.0.0/16)
            or CIDR 172.16/12 range.
            WARNING: this method DOES NOT check against multicast (class D) or
            reserved class E networks. If you need such test, you must do it
            manually (you can use 'multicast' and 'class_e' module attributes).
        """
        return (self in class_a) or (self in class_c) or (self in loopback) or (self in link_local)
        
    
    def is_loopback(self):
        """ Return True if current network belongs to or is part of loopback address range
            (127.0.0.0/8)
        """
        return self in loopback
    
    
    def is_link_local(self):
        """ Return True if current network belongs to or is part of link-local address range
            (169.254.0.0/16), which is used by some operating systems to assign an IP address
            to device, when there's is no configuration source or source is not available
            (i.e. configuration via DHCP fails).
        """
        return self in link_local
    
    
    def is_multicast(self):
        """ Return True if a network belongs or is a part of multicast range """
        return self in multicast
    
    
    def calcmask(self):
        """ Calculate the smallest netmask for the range """
        xor = self.first.numeric ^ self.last.numeric
        for i in xrange(32):
            if xor & ~(-1 << i) == xor: break
        return 32 - i
    
    
    def next(self):
        if self.__dict__['__iter_idx'] > self.last.numeric:
            self.reset()
            raise StopIteration
        v = self.__dict__['__iter_idx']
        self.__dict__['__iter_idx'] += 1
        return Address(v)
    
    
    def reset(self):
        """ Reset iteration index
            You shouldn't need it.
        """
        self.__dict__['__iter_idx'] = self.first.numeric
        
        
    def merge(self, other):
        """ Merge two Network objects together """
        if not isinstance(other, Network):
            raise ValueError('Cannot merge \'{0}\' with \'{1}\''.format(self.__class__.__name__, other.__class__.__name__))
        
        (net_a, net_b) = self._sort_networks(self, other)
        if net_b is None:
            # networks are the same
            return [Network(net_a),]
        if net_a.last < net_b.first:
            # there's a gap between networks, cannot merge
            return [net_a, net_b]
        
        return [Network((net_a.first, net_b.last)),]
    
    
    def intersect(self, other):
        """ Intersect two networks """
        if not isinstance(other, Network):
            raise ValueError('Cannot intersect \'{0}\' with \'{1}\''.format(self.__class__.__name__, other.__class__.__name__))
        
        (net_a, net_b) = self._sort_networks(self, other)
        if net_b is None or net_a.last == net_b.first - 1 or net_a.last < net_b.first:
            # [  net_a  ]   or   [  net_a  ]                or   [  net_a  ]
            # [  net_b  ]                   [  net_b  ]                       [  net_b  ]
            return [Network(net_a), Network(net_b)]
        if net_a.first < net_b.first and net_a.last < net_b.last:
            # [  net_a  ]
            #      [  net_b  ]
            return [Network((net_a.first, net_b.first)),
                    Network((net_b.first + 1, net_a.last)),
                    Network((net_a.last + 1, net_b.last))]
        elif net_a.first < net_b.first and net_a.last > net_b.last:
            # [      net_a     ]
            #    [  net_b  ]
            return [Network((net_a.first, net_b.first + 1)),
                    Network((net_b.last - 1, net_b.first)),
                    Network((net_b.first + 1, net_a.last))]
        elif net_a.first == net_b.first and net_a.last > net_b.last:
            # [    net_a    ]
            # [  net_b  ]
            return [Network((net_a.first, net_b.last)),
                    Network((net_b.first + 1, net_a.last))]
        elif net_a.first < net_b.first and net_a.last == net_b.last:
            # [    net_a    ]
            #     [  net_b  ]
            return [Network((net_a.first, net_b.first)),
                    Network((net_b.first + 1, net_b.last))]

        
    def _sort_networks(self, net_a, net_b):
        if net_a == net_b:
            # networks are the same
            return (net_a, None)
        elif net_a.first > net_b.first:
            # net_b is in front
            return (net_b, net_a)
        else:
            # net_a is in front
            return (net_a, net_b)
        

    def _parse_string(self, string):
        if string.find('/') >= 0:
            # 'network/mask'
            (address, mask) = string.split('/', 1)
            mask = self._parse_mask(mask)
            first = Address(Address.ston(address) & (((1<<mask) - 1)<<(32 - mask)))
            last = Address(first + (2**(32 - mask)) - 1)
        elif string.find('-') >= 0:
            # 'first - last' or 'first-last' (or even ' first   - last ')
            chunks = string.strip().replace(' ', '').split('-')
            if len(chunks) != 2:
                first = last = None
            else:
                (first, last) = map(Address, chunks)
            mask = None     # calculate
        else:
            first = last = mask = None
            
        if first == last == mask == None:
            # String passed to method doesn't look right
            raise ValueError('Invalid network value')
        
        return (first, last, mask)
    
    
    def _parse_address(self, value):
        if isinstance(value, (Address, str, unicode)):
            return Address(value)
        elif isinstance(value, (int, long)):
            return Address(int(value))
        else:
            raise ValueError('Don\'t know how to parse "{0}"'.format(value))
        
        
    def _parse_mask(self, value):
        if value == None:
            # calculate
            return value
        
        if isinstance(value, (str, unicode)):
            try:
                value = int(value)
            except:
                raise ValueError('Invalid netmask value')
        if value < 0 or value > 32:
            raise ValueError('Netmask value out of bounds')
        return value
    
    
    def __setattr__(self, name, value):
        if name in ('first', 'last'):
            self.__dict__['_' + name] = self._parse_address(value)
            self.__dict__['_mask'] = None
        elif name == 'mask':
            self.__dict__['_mask'] = self._parse_mask(value)
            if not value and not self.last:
                self.last = Address(self.first)
            elif value:
                self.__dict__['_last'] = Address(self.__dict__['_first'] + (2**(32 - self.mask)) - 1)
        elif name == 'size':
            raise AttributeError('Attribute \'{0}\' is read-only'.format(name))
        else:
            raise AttributeError('\'{0}\' object has no attribute \'{1}\''.format(self.__class__.__name__, name))
        
        
    def __getattr__(self, name):
        if name in ('first', 'last'):
            return self.__dict__['_' + name]
        elif name == 'mask':
            return self.__dict__['_mask'] or self.calcmask()
        elif name == 'size':
            return self.__len__()
        else:
            raise AttributeError('\'{0}\' object has no attribute \'{1}\''.format(self.__class__.__name__, name))
    
    
    def __iter__(self):
        return self
    
    
    def __len__(self):
        return self.last.numeric - self.first.numeric
    
    
    def __getitem__(self, srange):
        if isinstance(srange, slice):
            # return address subrange
            if srange.step is not None:
                raise IndexError('Stepping on \'{0}\' is not supported'.format(self.__class__.__name__))
            start = (self.first if srange.start > 0 else self.last) + srange.start
            stop = (self.first if srange.stop > 0 else self.last) + srange.stop
            return self.__class__(sorted((start, stop)))
        
        # return single IP address
        if srange > self.size or (srange * -1) > self.size:
            raise IndexError('Index out of bounds')
        
        if srange >= 0:
            return Address(self.first + srange)
        else:
            return Address(self.last + srange)
        
    
    def __str__(self):
        return '{0} - {1}'.format(self.first.literal, self.last.literal)
    
    
    def __repr__(self):
        return '{0}(\'{1} - {2}\')'.format(self.__class__.__name__, self.first.literal, self.last.literal)
        
    
    def __contains__(self, other):
        if isinstance(other, Network):
            return (other.first.numeric <= self.last.numeric and other.last.numeric >= self.last.numeric) or \
                   (other.last.numeric >= self.first.numeric and other.first.numeric <= self.first.numeric)
        elif isinstance(other, Address):
            return other.numeric >= self.first.numeric and other.numeric <= self.last.numeric
        else:
            raise ValueError('Cannot compare {0} with {1}'.format(self.__class__.__name__, other.__class__.__name))
        
        
    def __add__(self, value):
        return self.__class__((self.first, self.first + value))
    
    
    def __iadd__(self, value):
        self.last += value
        return self
    
    
    def __sub__(self, value):
        if value >= self.size:
            raise ValueError('Substracted value is greater than network size')
        return self.__class__((self.first, self.last - value))
    
    
    def __isub__(self, value):
        if value >= self.size:
            raise ValueError('Substracted value is greater than network size')
        self.last -= value
        return self
    
    
    def __eq__(self, other):
        if not isinstance(other, Network):
            raise ValueError('Cannot compare {0} with {1}'.format(self.__class__.__name__, other.__class__.__name__))
        return (self.first == other.first) and (self.last == other.last)
    
    
    def __ne__(self, other):
        return not self.__eq__(other)
    
    



class_a = Network('10.0.0.0 - 10.255.255.255')
class_b = Network('128.0.0.0 - 191.255.255.255')
class_c = Network('192.168.0.0 - 192.168.255.255')
class_d = Network('224.0.0.0 - 239.255.255.255')
class_e = Network('240.0.0.0 - 255.255.255.255')
cidr_172 = Network('172.16.0.0 - 172.31.255.255')
loopback = Network('127.0.0.0 - 127.255.255.255')
link_local = Network('169.254.0.0 - 169.254.255.255') # IPv4LL, APIPA
multicast = class_d
